// -*- C++ -*-
//==============================================================================================
//
//	This file is part of LiDIA --- a library for computational number theory
//
//	Copyright (c) 1994--2001 the LiDIA Group.  All rights reserved.
//
//	See http://www.informatik.tu-darmstadt.de/TI/LiDIA/
//
//----------------------------------------------------------------------------------------------
//
//	$Id$
//
//	Author	: Stefan Neis (SN)
//	Changes	: See CVS log
//
//==============================================================================================

//
// This  include  file  supports polynomial
// operations for all specializations.
//


#ifndef LIDIA_POLYNOMIAL_H_GUARD_
#define LIDIA_POLYNOMIAL_H_GUARD_



#ifndef LIDIA_POLY_INTERN_H_GUARD_
# include	"LiDIA/base/poly_intern.h"
#endif
#ifndef LIDIA_FIELD_POLYNOMIAL_H_GUARD_
# include	"LiDIA/field_polynomial.h"
#endif



#define LIDIA_CLASS_POLYNOMIAL

#include	"LiDIA/specialization/polynomial.special"



#endif	// LIDIA_POLYNOMIAL_H_GUARD_
