/* include/LiDIA/path.h.  Generated from path.h.in by configure.  */
// -*- C++ -*-
//==============================================================================================
//
//	This file is part of LiDIA --- a library for computational number theory
//
//	Copyright (c) 1994--2001 the LiDIA Group.  All rights reserved.
//
//	See http://www.informatik.tu-darmstadt.de/TI/LiDIA/
//
//----------------------------------------------------------------------------------------------
//
//	$Id$
//
//	Author	: C. Cornelssen
//	Changes	: See CVS log
//
//==============================================================================================


#ifndef LIDIA_PATH_H_GUARD_
#define LIDIA_PATH_H_GUARD_

/* The filename of LiDIA's GF(2^n) database */
#define LIDIA_GF2N_DB "/usr/share/lidia/GF2n.database"

/* The filename of LiDIA's primes list for remainder arithmetic */
#define LIDIA_PRIMELIST_NAME "/usr/share/lidia/LIDIA_PRIMES"

/* The directory holding LiDIA ECO's GF(p) database files */
#define LIDIA_ECO_DB "/usr/share/lidia/MOD_EQ"

/* The directory holding LiDIA ECO's GF(2^n) database files */
#define LIDIA_ECO2_DB "/usr/share/lidia/MOD_EQ_2"

/* The directory holding LiDIA GEC's imaginary quadratic discriminant database files */
#define LIDIA_DISCRIMINANTS_DB "/usr/share/lidia/Discriminants"

/* The file holding the differences of subsequential primes used for LiDIA GEC's prime_proo */
#define LIDIA_PRIME_DIFFS "/usr/share/lidia/LIDIA_PRIME_DIFFS"

#endif	// LIDIA_PATH_H_GUARD_

